c
c     stereo.f - Multibaseline stereo
c
c     CMU task parallel program suite
c
c     Peter Dinda, Thomas Gross, David O'Hallaron, Edward Segall,
c     James Stichnoth, Jaspal Subhlok, Jon Webb, Bwolen Yang
c
c     Okutomi/Kanade multiple baseline stereo using John Webb's columnn
c     summing scheme, adapted from J. Webb, "Implementation and performance
c     of fast parallel multi-baseline stereo vision", in Computing 
c     Architectures for Machine Perception, New Orleans, LA, Dec, 1993.
c
c     Note: Despite the parameters WINX and WINY, there are places in 
c     this code where a 13x13 window is assumed.
c
c     Comments and suggestions to David O'Hallaron, droh@cs.cmu.edu

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c     Copyright (c) 1994 by Carnegie Mellon University
c
c     Permission to use, copy, modify, and distribute this software
c     for any purpose and without fee is hereby granted, provided that 
c     the above copyright notice appear in all copies.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      program stereo

      include 'stereo.h'

c     input images
      integer ref(ROWS+WINY,COLS)     ! reference image
      integer m1(ROWS+WINY,COLS)      ! first match image
      integer m2(ROWS+WINY,COLS)      ! second match image

c     intermediate images
      real diffimg(ROWS+WINY,COLS)
      real curbesterr(ROWS,COLS)

c     output disparity image
      integer curbestdisp(ROWS,COLS)

c     misc
      integer count, curdisp
      real start 
      real dsecs,esecs,seconds

c
c     start the stereo program
c
      dsecs = 0.0
      esecs = 0.0
      do count=1,ITERS
         call getdata(ref,m1,m2,curbesterr,curbestdisp)
         start = seconds()
         do curdisp=0,MAXDISP-1
            call gendiffimg(ref, m1, m2, diffimg, curdisp)
            dsecs = dsecs + seconds()
            call updatedispimg(diffimg, curbesterr, 
     $           curbestdisp, curdisp)
            esecs = esecs + seconds()
         enddo
         call testdata(curbestdisp)
c         call dumpimg(curbestdisp,ROWS,COLS)
      enddo
      call prperf(dsecs,esecs,iters)
      end

c
c     prperf - print the performance numbers
c
      subroutine prperf(dsecs,esecs,iter)
      include 'stereo.h'
      real dsecs,esecs,secs
      integer dpercent,epercent,iter

      dsecs = dsecs/iter
      esecs = esecs/iter
      secs = dsecs+esecs
      
      dpercent = (dsecs/secs)*100
      epercent = (esecs/secs)*100
      
      print '(''******************************'')'
      print '(''Multibaseline stereo ('',i3,'' x '',i3,'')'')',
     $     ROWS, COLS
      print '(''******************************'')'
      print '(''Time for one image set:'')'
      print '(''diff     : '',f9.6,'' ('',i2,''%)'')',dsecs,dpercent
      print '(''error    : '',f9.6,'' ('',i2,''%)'')',esecs,epercent
      print '(''total(s) : '',f9.6)',secs
      print *, " "
      print '(''throughput : '',f6.2, '' frames/sec'')', 
     $     1.0/(dsecs+esecs)
      print *," "
      end
c
c     gendiffimg - this routine forms the difference image  
c
c
      subroutine gendiffimg(ref, m1, m2, diffimg, curdisp) 
      include 'stereo.h'

      integer ref(ROWS+WINY,COLS), m1(ROWS+WINY,COLS) 
      integer m2(ROWS+WINY,COLS)
      real diffimg(ROWS+WINY,COLS)
      integer curdisp

      integer i, j
      
      do i=1,ROWS+WINY
         do j=1,COLS-2*curdisp
            diffimg(i,j)=(ref(i,j)-m1(i,j+curdisp))**2 +
     $                   (ref(i,j)-m2(i,j+2*curdisp))**2
         enddo
      enddo

c      print *,"ref(124,109)=",ref(124,109)

      end


c 
c updatedispimg - This routine sums the difference image over a moving 
c                 window and updates the current best error and best 
c                 disparity images.
c
      subroutine updatedispimg(diffimg, curbesterr, 
     $                         curbestdisp, curdisp)
      include 'stereo.h'

      real diffimg(ROWS+WINY,COLS)
      real curbesterr(ROWS,COLS)
      integer curbestdisp(ROWS,COLS)
      integer curdisp

      real sum

      real csum(COLS)

      integer i, j

c      integer local

      real localf
      
c      do i=118,131
c         do j=1, COLS
c            local=diffimg(i,j)
c            print *,"diffimg(",i,",",j,")=",local
c        enddo
c      enddo


c      print *, "diffimg(124,109)=",diffimg(124,109)

      do j=1,COLS
         csum(j) = 0.0
         do i=1,WINX
            csum(j) = csum(j) + diffimg(i,j)
         enddo
      enddo

      do i=1,ROWS
         sum = 0.0
         do j=1,7
            sum = sum + csum(j)
         enddo
         
         if (sum .lt. curbesterr(i,1)) then
            curbesterr(i,1) = sum
            curbestdisp(i,1) = curdisp
         endif
         
         do j=2,7
            sum = sum + csum(j+6)
            if (sum .lt. curbesterr(i,j)) then
               curbesterr(i,j) = sum
               curbestdisp(i,j) = curdisp
            endif
         enddo
         
         do j=8,COLS-6
            sum = sum - csum(j-7)
            sum = sum + csum(j+6)
            if (sum .lt. curbesterr(i,j)) then
               curbesterr(i,j) = sum
               curbestdisp(i,j) = curdisp
            endif
         enddo
         
         do j=COLS-5,COLS
            sum = sum - csum(j-7)
            if (sum .lt. curbesterr(i,j)) then
               curbesterr(i,j) = sum
               curbestdisp(i,j) = curdisp
            endif
         enddo
         
         do j=1,COLS
            csum(j) = csum(j) - diffimg(i,j) + diffimg(i+WINY,j)
         enddo
         
      Enddo

      do i=118,131
         do j=1, COLS
            localf=curbesterr(i,j)
c            print *,"curbesterr(",i,",",j,")=",localf
        enddo
      enddo


c      print *, "curdisp=",curdisp

c      print *, "curbesterr(124,109)=",curbesterr(124,109)
c      print *, "curbestdisp(124,109)=",curbestdisp(124,109)

      end
      

C
C Generate some input data
C
C This routine is pretty dull.
C
      subroutine getdata(ref,m1,m2,curbesterr,curbestdisp)
      include 'stereo.h'

      integer tardis,isize,jsize
      parameter (tardis=12)
      parameter (isize=37)
      parameter (jsize=37)

      integer i, j

      integer ref(ROWS+WINY,COLS), m1(ROWS+WINY,COLS)
      integer m2(ROWS+WINY,COLS)
      real curbesterr(ROWS,COLS)
      integer curbestdisp(ROWS,COLS)

      integer ipos,jpos
      
      do i=1,6
         do j=1,COLS
            ref(i,j)=0
            m1(i,j)=0
            m2(i,j)=0
         enddo
      enddo


      do i=7,ROWS+6
         do j=1,COLS
            ref(i,j) = (i + j)
            m1(i,j) = (i + j)
            m2(i,j) = (i + j)
         enddo
      enddo


      do i=ROWS+7,ROWS+13
         do j=1,COLS
            ref(i,j)=0
            m1(i,j)=0
            m2(i,j)=0
         enddo
      enddo


      ipos=(ROWS-isize)/2 + 6
      jpos=((COLS-jsize)/2) - tardis
      
      do i=ipos,ipos+isize-1
         do j=jpos,jpos+jsize-1
            ref(i,j) = ((i-ipos+1) + (j-jpos+1))
            m1(i,j+tardis) = ((i-ipos+1) + (j-jpos+1))
            m2(i,j+2*tardis) = ((i-ipos+1) + (j-jpos+1))
         enddo
      enddo
      
      do i=1,ROWS
         do j=1,COLS
            curbesterr(i,j) = 9999999.0
            curbestdisp(i,j) = 0
         enddo
      enddo
      
      
      end


c
c Routine to test output (matched with input routine above)
c
c
      subroutine testdata(disp)
      include 'stereo.h'

      integer tardis,isize,jsize
      parameter (tardis=12)
      parameter (isize=37)
      parameter (jsize=37)

      integer disp(ROWS,COLS)

      integer ipos, jpos, i, j

      ipos=(ROWS-isize)/2
      jpos=((COLS-jsize)/2) - tardis
      
      do i=ipos+10,ipos+isize-11
         do j=jpos+10,jpos+jsize-11
            if (disp(i,j).ne.tardis) then
               write (*,100) i,j,tardis,disp(i,j)
            end if
         end do
      end do
      
      do i=1,ipos-10
         do j=1,jpos+jsize+2*tardis+9
            if (disp(i,j).ne.0) then
               write (*,100) i,j,0,disp(i,j)
            end if
         end do
      end do
      
      do i=1,ipos+isize+9
         do j=jpos+jsize+2*tardis+9,COLS
            if (disp(i,j).ne.0) then
               write (*,100) i,j,0,disp(i,j)
            end if
         end do
      end do
      
      do i=ipos+isize+9,ROWS
         do j=jpos-10,COLS
            if (disp(i,j).ne.0) then
               write (*,100) i,j,0,disp(i,j)
            end if
         end do
      end do
      
      do i=ipos-10,ROWS
         do j=1,jpos-10
            if (disp(i,j).ne.0) then
               write (*,100) i,j,0,disp(i,j)
            end if
         end do
      end do

 100  format("Error at position (",I3,",",I3,") - expected disp ",I,
     $       " but got ",I)

      
      end

      subroutine dumpimg(img,ROWS,COLS)

      integer rows, cols
      integer img(rows,cols)

      integer i,j
      
      print *,"#row col val"

      do i=1, ROWS
         do j=1,COLS
            write (*,200) i,j,img(i,j)
         enddo
      enddo

 200  format(I3," ",I3," ", I)

      end
            
c
c     seconds - returns elapsed seconds since last call to seconds().
c               the first time, returns elapsed time since program started. 
c
      real function seconds()
      real tarray(2),dtime
      seconds = dtime(tarray)
      seconds = tarray(1)
      end








